/*
 * Copyright (C) 2023 Red Hat
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"

#include "backends/native/meta-kms-lease.h"

#include <glib.h>

#include "backends/native/meta-crtc-kms.h"
#include "backends/native/meta-kms.h"
#include "backends/native/meta-kms-connector.h"
#include "backends/native/meta-kms-crtc-private.h"
#include "backends/native/meta-kms-device.h"
#include "backends/native/meta-kms-plane.h"

enum
{
  PROP_0,

  PROP_MANAGER_BACKEND_NATIVE,

  N_PROPS_MANAGER,
};

static GParamSpec *props_manager[N_PROPS_MANAGER] = { NULL };

enum
{
  MANAGER_CONNECTORS_CHANGED,

  N_SIGNALS_MANAGER,
};

static guint signals_manager[N_SIGNALS_MANAGER] = { 0 };

enum
{
  LEASE_REVOKED,

  N_SIGNALS_LEASE,
};

static guint signals_lease[N_SIGNALS_LEASE] = { 0 };

struct _MetaKmsLeaseManager
{
  GObject parent;

  MetaBackendNative *backend_native;

  gulong device_added_handler_id;
  gulong resources_changed_handler_id;
  gulong lease_changed_handler_id;

  GHashTable *leases;
  GHashTable *connectors;
};

G_DEFINE_TYPE (MetaKmsLeaseManager, meta_kms_lease_manager, G_TYPE_OBJECT)

typedef struct _LeasingKmsAssignment
{
  MetaKmsConnector *connector;
  MetaKmsCrtc *crtc;
  MetaKmsPlane *primary_plane;
  MetaKmsPlane *cursor_plane;
} LeasingKmsAssignment;

struct _MetaKmsLease
{
  GObject parent;

  uint32_t lessee_id;
  int fd;
  MetaKmsDevice *kms_device;
  GList *assignments;
};

G_DEFINE_TYPE (MetaKmsLease, meta_kms_lease, G_TYPE_OBJECT)

static MetaKmsCrtc *
find_crtc_to_lease (MetaKmsConnector *kms_connector)
{
  MetaKmsDevice *device = meta_kms_connector_get_device (kms_connector);
  const MetaKmsConnectorState *connector_state =
    meta_kms_connector_get_current_state (kms_connector);
  GList *l;

  for (l = meta_kms_device_get_crtcs (device); l; l = l->next)
    {
      MetaKmsCrtc *kms_crtc = l->data;
      MetaCrtcKms *crtc_kms = meta_crtc_kms_from_kms_crtc (kms_crtc);
      uint32_t crtc_idx;

      if (meta_crtc_is_leased (META_CRTC (crtc_kms)))
        continue;

      if (meta_crtc_get_outputs (META_CRTC (crtc_kms)) != NULL)
        continue;

      crtc_idx = meta_kms_crtc_get_idx (kms_crtc);
      if (!(connector_state->common_possible_crtcs & (1 << crtc_idx)))
        continue;

      return kms_crtc;
    }

  return NULL;
}

static gboolean
is_plane_assigned (MetaKmsDevice *kms_device,
                   MetaKmsPlane  *kms_plane)
{
  GList *l;

  for (l = meta_kms_device_get_crtcs (kms_device); l; l = l->next)
    {
      MetaKmsCrtc *kms_crtc = l->data;
      MetaCrtcKms *crtc_kms = meta_crtc_kms_from_kms_crtc (kms_crtc);

      if (meta_crtc_kms_get_assigned_primary_plane (crtc_kms) == kms_plane)
        return TRUE;
    }

  return FALSE;
}

static MetaKmsPlane *
find_plane_to_lease (MetaKmsCrtc      *kms_crtc,
                     MetaKmsPlaneType  plane_type)
{
  MetaKmsDevice *kms_device = meta_kms_crtc_get_device (kms_crtc);
  GList *l;

  for (l = meta_kms_device_get_planes (kms_device); l; l = l->next)
    {
      MetaKmsPlane *kms_plane = l->data;

      if (meta_kms_plane_get_plane_type (kms_plane) != plane_type)
        continue;

      if (!meta_kms_plane_is_usable_with (kms_plane, kms_crtc))
        continue;

      if (is_plane_assigned (kms_device, kms_plane))
        continue;

      return kms_plane;
    }

  return NULL;
}

static MetaKmsDevice *
find_resources_to_lease (GList   *connectors,
                         GList  **out_assignments,
                         GList  **out_crtcs,
                         GList  **out_planes,
                         GError **error)
{
  MetaKmsDevice *kms_device = NULL;
  g_autoptr (GList) assignments = NULL;
  g_autoptr (GList) crtcs = NULL;
  g_autoptr (GList) planes = NULL;
  GList *l;

  if (connectors == NULL)
    {
      g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
                   "Cannot create lease without connectors");
      return NULL;
    }

  for (l = connectors; l; l = l->next)
    {
      MetaKmsConnector *connector = l->data;
      MetaKmsDevice *connector_device;

      connector_device =
        meta_kms_connector_get_device (connector);

      if (kms_device != NULL && kms_device != connector_device)
        {
          g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
                       "Cannot create lease on multiple devices");
          return NULL;
        }

      kms_device = connector_device;
    }

  for (l = connectors; l; l = l->next)
    {
      MetaKmsConnector *connector = l->data;
      LeasingKmsAssignment *assignment;
      MetaKmsCrtc *crtc;
      MetaKmsPlane *primary_plane;
      MetaKmsPlane *cursor_plane;

      crtc = find_crtc_to_lease (connector);
      if (!crtc)
        {
          g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
                       "Failed to find CRTC to lease with connector %u (%s)",
                       meta_kms_connector_get_id (connector),
                       meta_kms_device_get_path (kms_device));
          return NULL;
        }

      crtcs = g_list_append (crtcs, crtc);

      primary_plane = find_plane_to_lease (crtc, META_KMS_PLANE_TYPE_PRIMARY);
      if (!primary_plane)
        {
          g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED,
                       "Failed to find primary plane "
                       "to lease with connector %u (%s)",
                       meta_kms_connector_get_id (connector),
                       meta_kms_device_get_path (kms_device));
          return NULL;
        }

      planes = g_list_append (planes, primary_plane);

      cursor_plane = find_plane_to_lease (crtc, META_KMS_PLANE_TYPE_CURSOR);
      if (!cursor_plane)
        {
          g_warning ("Failed to find cursor plane "
                     "to lease with connector %u (%s)",
                     meta_kms_connector_get_id (connector),
                     meta_kms_device_get_path (kms_device));
        }
      else
        {
          planes = g_list_append (planes, cursor_plane);
        }

      assignment = g_new0 (LeasingKmsAssignment, 1);
      assignment->connector = connector;
      assignment->crtc = crtc;
      assignment->primary_plane = primary_plane;
      assignment->cursor_plane = cursor_plane;

      assignments = g_list_append (assignments, assignment);
    }

  *out_assignments = g_steal_pointer (&assignments);
  *out_crtcs = g_steal_pointer (&crtcs);
  *out_planes = g_steal_pointer (&planes);
  return kms_device;
}

uint32_t
meta_kms_lease_get_id (MetaKmsLease *lease)
{
  return lease->lessee_id;
}

int
meta_kms_lease_steal_fd (MetaKmsLease *lease)
{
  int fd = lease->fd;
  lease->fd = -1;
  return fd;
}

gboolean
meta_kms_lease_is_active (MetaKmsLease *lease)
{
  return lease->lessee_id != 0;
}

static void
meta_kms_lease_assign (MetaKmsLease *lease)
{
  GList *l;

  for (l = lease->assignments; l; l = l->next)
    {
      LeasingKmsAssignment *assignment = l->data;
      MetaCrtcKms *crtc_kms = meta_crtc_kms_from_kms_crtc (assignment->crtc);

      meta_kms_crtc_set_is_leased (assignment->crtc, TRUE);
      meta_crtc_kms_assign_plane (crtc_kms,
                                  assignment->primary_plane,
                                  assignment->cursor_plane);
    }
}

static void
meta_kms_lease_unassign (MetaKmsLease *lease)
{
  GList *l;

  for (l = lease->assignments; l; l = l->next)
    {
      LeasingKmsAssignment *assignment = l->data;
      MetaCrtcKms *crtc_kms = meta_crtc_kms_from_kms_crtc (assignment->crtc);

      meta_kms_crtc_set_is_leased (assignment->crtc, FALSE);
      meta_crtc_kms_assign_plane (crtc_kms, NULL, NULL);
    }
}

static void
mark_revoked (MetaKmsLease *lease)
{
  meta_kms_lease_unassign (lease);

  g_signal_emit (lease, signals_lease[LEASE_REVOKED], 0);
  lease->lessee_id = 0;
}

void
meta_kms_lease_revoke (MetaKmsLease *lease)
{
  g_autoptr (GError) error;

  if (!lease->lessee_id)
    return;

  if (!meta_kms_device_revoke_lease (lease->kms_device, lease->lessee_id, &error))
    {
      g_warning ("Failed to revoke DRM lease on %s: %s",
                 meta_kms_device_get_path (lease->kms_device),
                 error->message);
      return;
    }

  mark_revoked (lease);
}

static void
meta_kms_lease_disappeared (MetaKmsLease *lease)
{
  mark_revoked (lease);
}

static void
meta_kms_lease_dispose (GObject *object)
{
  MetaKmsLease *lease = META_KMS_LEASE (object);

  g_clear_object (&lease->kms_device);

  if (lease->assignments)
    {
      g_list_free_full (lease->assignments, g_free);
      lease->assignments = NULL;
    }

  G_OBJECT_CLASS (meta_kms_lease_parent_class)->dispose (object);
}

static void
meta_kms_lease_finalize (GObject *object)
{
  MetaKmsLease *lease = META_KMS_LEASE (object);

  close (lease->fd);

  G_OBJECT_CLASS (meta_kms_lease_parent_class)->finalize (object);
}

static void
meta_kms_lease_class_init (MetaKmsLeaseClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  object_class->dispose = meta_kms_lease_dispose;
  object_class->finalize = meta_kms_lease_finalize;

  signals_lease[LEASE_REVOKED] =
    g_signal_new ("revoked",
                  G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  0,
                  NULL, NULL, NULL,
                  G_TYPE_NONE, 0);
}

static void
meta_kms_lease_init (MetaKmsLease *lease)
{
}

static void
on_lease_revoked (MetaKmsLease        *lease,
                  MetaKmsLeaseManager *lease_manager)
{
  GHashTableIter iter;
  MetaKmsConnector *connector;
  MetaKmsLease *other_lease;

  g_signal_handlers_disconnect_by_func (lease,
                                        on_lease_revoked,
                                        lease_manager);

  g_hash_table_iter_init (&iter, lease_manager->connectors);
  while (g_hash_table_iter_next (&iter,
                                 (gpointer *)&connector,
                                 (gpointer *)&other_lease))
    {
      if (lease == other_lease)
        g_hash_table_insert (lease_manager->connectors, connector, NULL);
    }

  g_hash_table_remove (lease_manager->leases,
                       GUINT_TO_POINTER (lease->lessee_id));

  g_signal_emit (lease_manager, signals_manager[MANAGER_CONNECTORS_CHANGED], 0);
}

MetaKmsLease *
meta_kms_lease_manager_lease_connectors (MetaKmsLeaseManager  *lease_manager,
                                         GList                *connectors,
                                         GError              **error)
{
  MetaKmsLease *lease;
  g_autoptr (GList) assignments = NULL;
  g_autoptr (GList) crtcs = NULL;
  g_autoptr (GList) planes = NULL;
  MetaKmsDevice *kms_device = NULL;
  int fd;
  uint32_t lessee_id;
  GList *l;

  kms_device = find_resources_to_lease (connectors,
                                        &assignments,
                                        &crtcs,
                                        &planes,
                                        error);
  if (!kms_device)
    return NULL;

  if (!meta_kms_device_lease_objects (kms_device,
                                      connectors, crtcs, planes,
                                      &fd, &lessee_id,
                                      error))
    return NULL;

  // FIXME: do we need to ref all the objects?
  lease = g_object_new (META_TYPE_KMS_LEASE, NULL);
  lease->lessee_id = lessee_id;
  lease->fd = fd;
  lease->kms_device = g_object_ref (kms_device);
  lease->assignments = g_steal_pointer (&assignments);

  meta_kms_lease_assign (lease);

  g_signal_connect_after (lease, "revoked", G_CALLBACK (on_lease_revoked),
                          lease_manager);

  for (l = connectors; l; l = l->next)
    {
      MetaKmsConnector *connector = l->data;

      g_hash_table_insert (lease_manager->connectors,
                           connector, lease);
    }

  g_hash_table_insert (lease_manager->leases,
                       GUINT_TO_POINTER (lessee_id), g_object_ref (lease));

  g_signal_emit (lease_manager,
                 signals_manager[MANAGER_CONNECTORS_CHANGED], 0);

  return lease;
}

GList *
meta_kms_lease_manager_get_connectors (MetaKmsLeaseManager *lease_manager)
{
  return g_hash_table_get_keys (lease_manager->connectors);
}

MetaKmsConnector *
meta_kms_lease_manager_get_connector_from_id (MetaKmsLeaseManager *lease_manager,
                                              uint32_t             connector_id)
{
  GHashTableIter iter;
  MetaKmsConnector *connector;

  g_hash_table_iter_init (&iter, lease_manager->connectors);
  while (g_hash_table_iter_next (&iter, (gpointer *)&connector, NULL))
    {
      if (meta_kms_connector_get_id (connector) == connector_id)
        return connector;
    }

  return NULL;
}

MetaKmsLease *
meta_kms_lease_manager_get_lease_from_connector (MetaKmsLeaseManager *lease_manager,
                                                 MetaKmsConnector    *kms_connector)
{
  return g_hash_table_lookup (lease_manager->connectors, kms_connector);
}



MetaKmsLease *
meta_kms_lease_manager_get_lease_from_id (MetaKmsLeaseManager *lease_manager,
                                          uint32_t             lessee_id)
{
  return g_hash_table_lookup (lease_manager->leases,
                              GUINT_TO_POINTER (lessee_id));
}

static void
update_connectors (MetaKmsLeaseManager *lease_manager)
{
  MetaKms *kms = meta_backend_native_get_kms (lease_manager->backend_native);
  GHashTable *new_connectors;
  MetaKmsLease *lease;
  GList *l;
  GList *o;
  gboolean has_new = FALSE;

  new_connectors = g_hash_table_new_similar (lease_manager->connectors);

  for (l = meta_kms_get_devices (kms); l; l = l->next)
    {
      MetaKmsDevice *kms_device = l->data;

      for (o = meta_kms_device_get_connectors (kms_device); o; o = o->next)
        {
          MetaKmsConnector *kms_connector = o->data;

          if (!meta_kms_connector_is_for_lease (kms_connector))
            continue;

          if (!g_hash_table_steal_extended (lease_manager->connectors,
                                            kms_connector,
                                            NULL, (gpointer *) &lease))
            has_new = TRUE;
          g_hash_table_insert (new_connectors, kms_connector, lease);
        }
    }

  if (has_new || g_hash_table_size (lease_manager->connectors) != 0)
    {
      g_signal_emit (lease_manager,
                     signals_manager[MANAGER_CONNECTORS_CHANGED], 0);
    }

  g_clear_pointer (&lease_manager->connectors, g_hash_table_unref);
  lease_manager->connectors = new_connectors;
}

static void
lease_disappeared (MetaKmsLeaseManager *lease_manager,
                   MetaKmsLease        *lease)
{
  GList *l;

  for (l = lease->assignments; l; l = l->next)
    {
      LeasingKmsAssignment *assignment = l->data;
      MetaKmsConnector *kms_connector = assignment->connector;

      if (g_hash_table_lookup_extended (lease_manager->connectors,
                                        kms_connector,
                                        NULL, NULL))
        g_hash_table_insert (lease_manager->connectors, kms_connector, NULL);
    }

  meta_kms_lease_disappeared (lease);
}

static gboolean
did_lease_disappear (MetaKmsLease  *lease,
                     uint32_t      *lessees,
                     int            num_lessees,
                     MetaKmsDevice *kms_device)
{
  int i;

  if (lease->kms_device != kms_device)
    return FALSE;

  for (i = 0; i < num_lessees; i++)
    {
      if (lease->lessee_id == lessees[i])
        return FALSE;
    }

  return TRUE;
}

static void
update_leases (MetaKmsLeaseManager *lease_manager)
{
  MetaKms *kms = meta_backend_native_get_kms (lease_manager->backend_native);
  MetaKmsLease *lease;
  GList *l;
  g_autoptr (GList) disappeared_leases = NULL;

  for (l = meta_kms_get_devices (kms); l; l = l->next)
    {
      MetaKmsDevice *kms_device = l->data;
      g_autofree uint32_t *lessees = NULL;
      int num_lessees;
      g_autoptr (GError) error = NULL;
      GHashTableIter iter;

      if (!meta_kms_device_list_lessees (kms_device,
                                         &lessees, &num_lessees,
                                         &error))
        {
          g_warning ("Failed to list leases: %s", error->message);
          continue;
        }

      g_hash_table_iter_init (&iter, lease_manager->leases);
      while (g_hash_table_iter_next (&iter, NULL, (gpointer *)&lease))
        {
          if (did_lease_disappear (lease, lessees, num_lessees, kms_device))
            disappeared_leases = g_list_append (disappeared_leases, lease);
        }

      g_free (lessees);
    }

  for (l = disappeared_leases; l; l = l->next)
    {
      lease = l->data;

      lease_disappeared (lease_manager, lease);
    }
}

static void
on_device_added (MetaKms             *kms,
                 MetaKmsDevice       *kms_device,
                 MetaKmsLeaseManager *lease_manager)
{
  update_connectors (lease_manager);
}

static void
on_resources_changed (MetaKms                *kms,
                      MetaKmsResourceChanges  changes,
                      MetaKmsLeaseManager    *lease_manager)
{
  if (changes != META_KMS_RESOURCE_CHANGE_FULL)
    return;

  update_connectors (lease_manager);
}

static void
on_lease_changed (MetaKms             *kms,
                  MetaKmsLeaseManager *lease_manager)
{
  update_leases (lease_manager);
}

static void
meta_kms_lease_manager_constructed (GObject *object)
{
  MetaKmsLeaseManager *lease_manager = META_KMS_LEASE_MANAGER (object);
  MetaKms *kms = meta_backend_native_get_kms (lease_manager->backend_native);

  lease_manager->device_added_handler_id =
    g_signal_connect (kms, "device-added",
                      G_CALLBACK (on_device_added),
                      lease_manager);
  lease_manager->resources_changed_handler_id =
    g_signal_connect (kms, "resources-changed",
                      G_CALLBACK (on_resources_changed),
                      lease_manager);
  lease_manager->lease_changed_handler_id =
    g_signal_connect (kms, "lease-changed",
                      G_CALLBACK (on_lease_changed),
                      lease_manager);

  lease_manager->leases =
    g_hash_table_new_full (NULL, NULL,
                           NULL,
                           (GDestroyNotify) g_object_unref);

  lease_manager->connectors =
    g_hash_table_new_full (NULL, NULL,
                           NULL, NULL);

  update_connectors (lease_manager);

  G_OBJECT_CLASS (meta_kms_lease_manager_parent_class)->constructed (object);
}

static void
meta_kms_lease_manager_set_property (GObject      *object,
                                     guint         prop_id,
                                     const GValue *value,
                                     GParamSpec   *pspec)
{
  MetaKmsLeaseManager *lease_manager = META_KMS_LEASE_MANAGER (object);
  switch (prop_id)
    {
    case PROP_MANAGER_BACKEND_NATIVE:
      lease_manager->backend_native = g_value_get_object (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
meta_kms_lease_manager_get_property (GObject    *object,
                                     guint       prop_id,
                                     GValue     *value,
                                     GParamSpec *pspec)
{
  MetaKmsLeaseManager *lease_manager = META_KMS_LEASE_MANAGER (object);
  switch (prop_id)
    {
    case PROP_MANAGER_BACKEND_NATIVE:
      g_value_set_object (value, lease_manager->backend_native);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
meta_kms_lease_manager_dispose (GObject *object)
{
  MetaKmsLeaseManager *lease_manager = META_KMS_LEASE_MANAGER (object);
  MetaKms *kms = meta_backend_native_get_kms (lease_manager->backend_native);

  g_clear_signal_handler (&lease_manager->device_added_handler_id, kms);
  g_clear_signal_handler (&lease_manager->resources_changed_handler_id, kms);
  g_clear_signal_handler (&lease_manager->lease_changed_handler_id, kms);

  g_clear_pointer (&lease_manager->leases, g_hash_table_unref);
  g_clear_pointer (&lease_manager->connectors, g_hash_table_unref);

  G_OBJECT_CLASS (meta_kms_lease_manager_parent_class)->dispose (object);
}

static void
meta_kms_lease_manager_class_init (MetaKmsLeaseManagerClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  object_class->constructed = meta_kms_lease_manager_constructed;
  object_class->set_property = meta_kms_lease_manager_set_property;
  object_class->get_property = meta_kms_lease_manager_get_property;
  object_class->dispose = meta_kms_lease_manager_dispose;

  props_manager[PROP_MANAGER_BACKEND_NATIVE] =
    g_param_spec_object ("backend-native", NULL, NULL,
                         META_TYPE_BACKEND_NATIVE,
                         G_PARAM_READWRITE |
                         G_PARAM_CONSTRUCT_ONLY |
                         G_PARAM_STATIC_STRINGS);

  g_object_class_install_properties (object_class,
                                     N_PROPS_MANAGER, props_manager);

  signals_manager[MANAGER_CONNECTORS_CHANGED] =
    g_signal_new ("connectors-changed",
                  G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  0,
                  NULL, NULL, NULL,
                  G_TYPE_NONE, 0);
}

static void
meta_kms_lease_manager_init (MetaKmsLeaseManager *lease_manager)
{
}
