/*
 * Copyright (C) 2016 Red Hat Inc.
 * Copyright (C) 2017 Intel Corporation
 * Copyright (C) 2018,2019 DisplayLink (UK) Ltd.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 */

#include "config.h"

#include "wayland/meta-wayland-drm-lease.h"

#include <glib.h>

#include "backends/native/meta-backend-native.h"
#include "backends/native/meta-kms-connector.h"
#include "backends/native/meta-kms-device.h"
#include "backends/native/meta-kms-device-private.h"
#include "backends/native/meta-kms-impl-device.h"
#include "backends/native/meta-kms-lease.h"
#include "backends/native/meta-kms.h"
#include "backends/edid.h"
#include "wayland/meta-wayland-private.h"

#include "drm-lease-v1-server-protocol.h"

struct _MetaWaylandDrmLeaseManager
{
  MetaWaylandCompositor *compositor;
  MetaKmsLeaseManager *kms_lease_manager;

  /* Key:   MetaKmsDevice *kms_device
   * Value: MetaWaylandDrmLeaseDevice *lease_device
   */
  GHashTable *devices;

  GList *leases;
};

typedef struct _MetaWaylandDrmLeaseDevice
{
  MetaWaylandDrmLeaseManager *lease_manager;

  struct wl_global *global;
  MetaKmsDevice *kms_device;

  /* Key:   MetaKmsConnector *kms_connector
   * Value: MetaWaylandDrmLeaseConnector *lease_connector
   */
  GHashTable *connectors;

  GList *resources;
} MetaWaylandDrmLeaseDevice;

typedef struct _MetaWaylandDrmLeaseConnector
{
  MetaWaylandDrmLeaseDevice *lease_device;

  MetaKmsConnector *kms_connector;
  char *description;

  GList *resources;
} MetaWaylandDrmLeaseConnector;

typedef struct _MetaWaylandDrmLeaseRequest
{
  MetaWaylandDrmLeaseDevice *lease_device;
  GList *lease_connectors;
  struct wl_resource *resource;
} MetaWaylandDrmLeaseRequest;

typedef struct _MetaWaylandDrmLease
{
  MetaWaylandDrmLeaseManager *lease_manager;
  MetaWaylandDrmLeaseDevice *lease_device;
  GList *lease_connectors;
  uint32_t lessee_id;
  struct wl_resource *resource;
} MetaWaylandDrmLease;

static void
meta_wayland_drm_lease_device_free (MetaWaylandDrmLeaseDevice *lease_device)
{
  g_clear_pointer (&lease_device->connectors, g_hash_table_unref);
}

static void
meta_wayland_drm_lease_device_release (MetaWaylandDrmLeaseDevice *lease_device)
{
  g_rc_box_release_full (lease_device,
                         (GDestroyNotify) meta_wayland_drm_lease_device_free);
}

static void
meta_wayland_drm_lease_connector_free (MetaWaylandDrmLeaseConnector *lease_connector)
{
  g_free (lease_connector->description);
  meta_wayland_drm_lease_device_release (lease_connector->lease_device);
}

static void
meta_wayland_drm_lease_connector_release (MetaWaylandDrmLeaseConnector *lease_connector)
{
  g_rc_box_release_full (lease_connector,
                         (GDestroyNotify) meta_wayland_drm_lease_connector_free);
}

static void
meta_wayland_drm_lease_free (MetaWaylandDrmLease *lease)
{
  meta_wayland_drm_lease_device_release (lease->lease_device);
  g_list_foreach (lease->lease_connectors,
                  (GFunc) meta_wayland_drm_lease_connector_release,
                  NULL);
}

static void
meta_wayland_drm_lease_release (MetaWaylandDrmLease *lease)
{
  g_rc_box_release_full (lease, (GDestroyNotify) meta_wayland_drm_lease_free);
}

static gboolean
meta_wayland_drm_has_active_leases (GList *leases)
{
  GList *l;

  for (l = leases; l; l = l->next)
    {
      MetaWaylandDrmLease *lease = l->data;
      MetaKmsLease *kms_lease =
        meta_kms_lease_manager_get_lease_from_id (lease->lease_manager->kms_lease_manager,
                                                  lease->lessee_id);

      if (kms_lease && meta_kms_lease_is_active (kms_lease))
        return TRUE;
    }

  return FALSE;
}

static void
meta_wayland_drm_lease_revoke (MetaWaylandDrmLease *lease)
{
  MetaKmsLease *kms_lease =
    meta_kms_lease_manager_get_lease_from_id (lease->lease_manager->kms_lease_manager,
                                              lease->lessee_id);

  if (kms_lease)
    {
      meta_kms_lease_revoke (kms_lease);
      wp_drm_lease_v1_send_finished (lease->resource);
    }
}

static void
wp_drm_lease_destroy (struct wl_client   *client,
                      struct wl_resource *resource)
{
  MetaWaylandDrmLease *lease = wl_resource_get_user_data (resource);

  meta_wayland_drm_lease_revoke (lease);

  wl_resource_destroy (resource);
}

static const struct wp_drm_lease_v1_interface drm_lease_implementation = {
  wp_drm_lease_destroy,
};

static void
wp_drm_lease_destructor (struct wl_resource *resource)
{
  MetaWaylandDrmLease *lease = wl_resource_get_user_data (resource);

  lease->lease_manager->leases = g_list_remove (lease->lease_manager->leases,
                                                lease);
  meta_wayland_drm_lease_release (lease);
}

static void
wp_drm_lease_request_request_connector (struct wl_client   *client,
                                        struct wl_resource *resource,
                                        struct wl_resource *connector)
{
  MetaWaylandDrmLeaseRequest *lease_request =
    wl_resource_get_user_data (resource);
  MetaWaylandDrmLeaseConnector *lease_connector =
    wl_resource_get_user_data (connector);

  if (lease_request->lease_device != lease_connector->lease_device)
    {
      wl_resource_post_error (resource,
                              WP_DRM_LEASE_REQUEST_V1_ERROR_WRONG_DEVICE,
                              "Wrong lease device");
      return;
    }

  if (g_list_find (lease_request->lease_connectors, lease_connector))
    {
      wl_resource_post_error (resource,
                              WP_DRM_LEASE_REQUEST_V1_ERROR_DUPLICATE_CONNECTOR,
                              "Connector requested twice");
      return;
    }

  lease_request->lease_connectors =
    g_list_append (lease_request->lease_connectors,
                   g_rc_box_acquire (lease_connector));
}

static void
wp_drm_lease_request_submit (struct wl_client   *client,
                             struct wl_resource *resource,
                             uint32_t            id)
{
  MetaWaylandDrmLeaseRequest *lease_request =
    wl_resource_get_user_data (resource);
  MetaWaylandDrmLeaseDevice *lease_device = lease_request->lease_device;
  MetaWaylandDrmLeaseManager *lease_manager = lease_device->lease_manager;
  GHashTable *available_devices = lease_manager->devices;
  GHashTable *available_connectors = lease_device->connectors;
  MetaKmsDevice *kms_device = lease_device->kms_device;
  MetaKmsLeaseManager *kms_lease_manager = lease_manager->kms_lease_manager;
  MetaWaylandDrmLease *lease;
  g_autoptr (GList) connectors = NULL;
  g_autoptr (MetaKmsLease) kms_lease = NULL;
  g_autoptr (GError) error = NULL;
  int fd = -1;
  GList *l;

  if (!lease_request->lease_connectors)
    {
      wl_resource_post_error (resource,
                              WP_DRM_LEASE_REQUEST_V1_ERROR_EMPTY_LEASE,
                              "Empty DRM lease request");
      return;
    }

  lease = g_rc_box_new0 (MetaWaylandDrmLease);
  lease->lease_manager = lease_manager;
  lease->lease_device = g_rc_box_acquire (lease_device);
  lease->resource =
    wl_resource_create (client, &wp_drm_lease_v1_interface,
                        wl_resource_get_version (resource), id);

  wl_resource_set_implementation (lease->resource,
                                  &drm_lease_implementation,
                                  lease,
                                  wp_drm_lease_destructor);

  lease_manager->leases = g_list_append (lease_manager->leases, lease);

  if (meta_wayland_drm_has_active_leases (lease_manager->leases))
    {
      g_warning ("Another lease is active at the moment");
      goto reject_request;
    }

  if (!g_hash_table_contains (available_devices, kms_device))
    {
      g_warning ("KMS device %s is not available",
                 meta_kms_device_get_path (kms_device));
      goto reject_request;
    }

  for (l = lease_request->lease_connectors; l; l = l->next)
    {
      MetaWaylandDrmLeaseConnector *lease_connector = l->data;
      MetaKmsConnector *kms_connector = lease_connector->kms_connector;

      if (!g_hash_table_contains (available_connectors, kms_connector))
        {
          g_warning ("KMS connector %u (%s) is not available",
                     meta_kms_connector_get_id (kms_connector),
                     meta_kms_device_get_path (kms_device));
          goto reject_request;
        }

      lease->lease_connectors =
        g_list_append (lease->lease_connectors,
                       g_rc_box_acquire (lease_connector));
      connectors = g_list_append (connectors, kms_connector);
    }

  kms_lease = meta_kms_lease_manager_lease_connectors (kms_lease_manager,
                                                       connectors,
                                                       &error);
  if (!kms_lease)
    {
      g_warning ("Failed to create lease from connector list: %s",
                 error->message);
      goto reject_request;
    }

  fd = meta_kms_lease_steal_fd (kms_lease);
  wp_drm_lease_v1_send_lease_fd (lease->resource, fd);

  lease->lessee_id = meta_kms_lease_get_id (kms_lease);

  goto clean;

reject_request:
  wp_drm_lease_v1_send_finished (lease->resource);

clean:
  if (fd != -1)
    close (fd);

  wl_resource_destroy (resource);
}

static const struct wp_drm_lease_request_v1_interface drm_lease_request_implementation = {
  wp_drm_lease_request_request_connector,
  wp_drm_lease_request_submit,
};

static void
wp_drm_lease_request_destructor (struct wl_resource *resource)
{
  MetaWaylandDrmLeaseRequest *lease_request =
    wl_resource_get_user_data (resource);

  meta_wayland_drm_lease_device_release (lease_request->lease_device);
  g_list_foreach (lease_request->lease_connectors,
                  (GFunc) meta_wayland_drm_lease_connector_release,
                  NULL);
  g_free (lease_request);
}

static void
wp_drm_lease_device_create_lease_request (struct wl_client   *client,
                                          struct wl_resource *resource,
                                          uint32_t            id)
{
  MetaWaylandDrmLeaseDevice *lease_device =
    wl_resource_get_user_data (resource);
  MetaWaylandDrmLeaseRequest *lease_request;

  lease_request = g_new0 (MetaWaylandDrmLeaseRequest, 1);
  lease_request->lease_device = g_rc_box_acquire (lease_device);
  lease_request->resource =
    wl_resource_create (client, &wp_drm_lease_request_v1_interface,
                        wl_resource_get_version (resource), id);

  wl_resource_set_implementation (lease_request->resource,
                                  &drm_lease_request_implementation,
                                  lease_request,
                                  wp_drm_lease_request_destructor);
}

static void
wp_drm_lease_device_release (struct wl_client   *client,
                             struct wl_resource *resource)
{
  wp_drm_lease_device_v1_send_released (resource);
  wl_resource_destroy (resource);
}

static const struct wp_drm_lease_device_v1_interface drm_lease_device_implementation = {
  wp_drm_lease_device_create_lease_request,
  wp_drm_lease_device_release,
};

static MetaWaylandDrmLeaseConnector *
meta_wayland_drm_lease_connector_new (MetaWaylandDrmLeaseDevice *lease_device,
                                      MetaKmsConnector          *kms_connector)
{
  MetaWaylandDrmLeaseConnector *lease_connector;
  const MetaKmsConnectorState *connector_state;

  lease_connector = g_rc_box_new0 (MetaWaylandDrmLeaseConnector);
  lease_connector->lease_device = g_rc_box_acquire (lease_device);
  lease_connector->kms_connector = kms_connector;

  connector_state = meta_kms_connector_get_current_state (kms_connector);
  if (connector_state->edid_data)
    {
      gconstpointer edid_data;
      g_autofree MetaEdidInfo *edid_info = NULL;
      size_t edid_size;
      g_autofree char *vendor = NULL;
      g_autofree char *product = NULL;
      GString *description;

      edid_data = g_bytes_get_data (connector_state->edid_data, &edid_size);
      edid_info = meta_edid_info_new_parse (edid_data, edid_size);

      description = g_string_new (NULL);

      vendor = g_strndup (edid_info->manufacturer_code, 4);
      if (vendor && g_utf8_validate (vendor, -1, NULL))
        g_string_append_printf (description, "%s", vendor);

      product = g_strndup (edid_info->dsc_product_name, 14);
      if (product && g_utf8_validate (product, -1, NULL))
        {
          if (description->len > 0)
            g_string_append_c (description, ' ');
          g_string_append_printf (description, "%s", product);
        }

      if (description->len == 0)
        {
          g_string_append_printf (description, "%s",
                                  meta_kms_connector_get_name (kms_connector));
        }

      lease_connector->description = g_string_free (description, FALSE);
    }
  else
    {
      lease_connector->description = g_strdup ("");
    }

  return lease_connector;
}

static void
meta_wayland_drm_lease_connector_send_withdrawn (MetaWaylandDrmLeaseConnector *lease_connector)
{
  GList *l;

  for (l = lease_connector->resources; l; l = l->next)
    {
      struct wl_resource *resource = l->data;

      if (wl_resource_get_user_data (resource) == lease_connector)
        wp_drm_lease_connector_v1_send_withdrawn (resource);
    }
}

static void
drm_lease_connector_destroy (struct wl_client   *client,
                             struct wl_resource *resource)
{
  wl_resource_destroy (resource);
}

static const struct wp_drm_lease_connector_v1_interface drm_lease_connector_implementation = {
  drm_lease_connector_destroy,
};

static void
wp_drm_lease_connector_destructor (struct wl_resource *resource)
{
  MetaWaylandDrmLeaseConnector *lease_connector =
    wl_resource_get_user_data (resource);

  lease_connector->resources = g_list_remove (lease_connector->resources,
                                              resource);
  meta_wayland_drm_lease_connector_release (lease_connector);
}

static void
send_new_connector_resource (MetaWaylandDrmLeaseDevice    *lease_device,
                             struct wl_resource           *device_resource,
                             MetaWaylandDrmLeaseConnector *lease_connector)
{
  struct wl_resource *connector_resource;
  const char *connector_name;
  uint32_t connector_id;

  connector_resource =
    wl_resource_create (wl_resource_get_client (device_resource),
                        &wp_drm_lease_connector_v1_interface,
                        wl_resource_get_version (device_resource),
                        0);
  wl_resource_set_implementation (connector_resource,
                                  &drm_lease_connector_implementation,
                                  g_rc_box_acquire (lease_connector),
                                  wp_drm_lease_connector_destructor);

  lease_connector->resources = g_list_append (lease_connector->resources,
                                              connector_resource);

  connector_name = meta_kms_connector_get_name (lease_connector->kms_connector);
  connector_id = meta_kms_connector_get_id (lease_connector->kms_connector);

  wp_drm_lease_device_v1_send_connector (device_resource, connector_resource);
  wp_drm_lease_connector_v1_send_name (connector_resource, connector_name);
  wp_drm_lease_connector_v1_send_description (connector_resource,
                                              lease_connector->description);
  wp_drm_lease_connector_v1_send_connector_id (connector_resource,
                                               connector_id);
  wp_drm_lease_connector_v1_send_done (connector_resource);
}

static void
send_connectors (MetaWaylandDrmLeaseDevice *lease_device,
                 struct wl_resource        *device_resource)
{
  GHashTableIter iter;
  MetaWaylandDrmLeaseConnector *lease_connector;

  g_hash_table_iter_init (&iter, lease_device->connectors);
  while (g_hash_table_iter_next (&iter, NULL, (gpointer *) &lease_connector))
    send_new_connector_resource (lease_device, device_resource, lease_connector);
}

static void
send_drm_fd (struct wl_client          *client,
             MetaWaylandDrmLeaseDevice *lease_device,
             struct wl_resource        *device_resource)
{
  int fd;
  MetaKmsImplDevice *impl_device;

  impl_device = meta_kms_device_get_impl_device (lease_device->kms_device);
  fd = meta_kms_impl_device_get_non_master_fd (impl_device);
  if (fd < 0)
    {
      wl_client_post_implementation_error (client,
                                           "Error getting DRM lease device fd");
      return;
    }

  wp_drm_lease_device_v1_send_drm_fd (device_resource, fd);
  close (fd);
}

static void
wp_drm_lease_device_destructor (struct wl_resource *resource)
{
  MetaWaylandDrmLeaseDevice *lease_device =
    wl_resource_get_user_data (resource);

  lease_device->resources = g_list_remove (lease_device->resources, resource);
  meta_wayland_drm_lease_device_release (lease_device);
}

static void
lease_device_bind (struct wl_client *client,
                   void             *user_data,
                   uint32_t          version,
                   uint32_t          id)
{
  MetaWaylandDrmLeaseDevice *lease_device = user_data;
  struct wl_resource *resource;

  resource = wl_resource_create (client, &wp_drm_lease_device_v1_interface,
                                 version, id);
  wl_resource_set_implementation (resource,
                                  &drm_lease_device_implementation,
                                  g_rc_box_acquire (lease_device),
                                  wp_drm_lease_device_destructor);

  send_drm_fd (client, lease_device, resource);
  send_connectors (lease_device, resource);
  wp_drm_lease_device_v1_send_done (resource);

  lease_device->resources = g_list_prepend (lease_device->resources, resource);
}

static void
meta_wayland_drm_lease_device_add_connector (MetaKmsConnector          *kms_connector,
                                             MetaWaylandDrmLeaseDevice *lease_device)
{
  MetaWaylandDrmLeaseConnector *lease_connector;

  if (!meta_kms_connector_is_for_lease (kms_connector))
    return;

  lease_connector = meta_wayland_drm_lease_connector_new (lease_device,
                                                          kms_connector);
  g_hash_table_insert (lease_device->connectors,
                       kms_connector,
                       g_steal_pointer (&lease_connector));
}

static void
meta_wayland_drm_lease_device_update_connectors (MetaWaylandDrmLeaseDevice *lease_device)
{
  GList *kms_connectors;
  GList *l;
  GHashTable *new_connectors;
  MetaWaylandDrmLeaseConnector *lease_connector;
  GHashTableIter iter;
  gboolean needs_done = FALSE;

  new_connectors = g_hash_table_new_similar (lease_device->connectors);

  kms_connectors = meta_kms_device_get_connectors (lease_device->kms_device);
  for (l = kms_connectors; l; l = l->next)
    {
      MetaKmsConnector *kms_connector = l->data;

      if (!meta_kms_connector_is_for_lease (kms_connector))
        continue;

      if (!g_hash_table_steal_extended (lease_device->connectors,
                                        kms_connector,
                                        NULL,
                                        (gpointer *) &lease_connector))
        {
          GList *lc;

          lease_connector =
            meta_wayland_drm_lease_connector_new (lease_device, kms_connector);

          for (lc = lease_device->resources; lc; lc = lc->next)
            {
              struct wl_resource *resource = lc->data;

              if (wl_resource_get_user_data (resource) == lease_device)
                {
                  send_new_connector_resource (lease_device,
                                               resource,
                                               lease_connector);
                  needs_done = TRUE;
                }
            }
        }

      g_hash_table_insert (new_connectors, kms_connector, lease_connector);
    }

  g_hash_table_iter_init (&iter, lease_device->connectors);
  while (g_hash_table_iter_next (&iter, NULL, (gpointer *) &lease_connector))
    {
      needs_done = TRUE;

      for (l = lease_device->lease_manager->leases; l; l = l->next)
        {
          MetaWaylandDrmLease *lease = l->data;

          if (g_list_find (lease->lease_connectors, lease_connector))
            meta_wayland_drm_lease_revoke (lease);
        }

      meta_wayland_drm_lease_connector_send_withdrawn (lease_connector);
    }

  if (needs_done)
    {
      g_list_foreach (lease_device->resources,
                      (GFunc) wp_drm_lease_device_v1_send_done,
                      NULL);
    }

  g_hash_table_unref (lease_device->connectors);
  lease_device->connectors = new_connectors;
}

static MetaWaylandDrmLeaseDevice *
meta_wayland_drm_lease_device_new (MetaWaylandDrmLeaseManager *lease_manager,
                                   MetaKmsDevice              *kms_device)
{
  struct wl_display *wayland_display =
    meta_wayland_compositor_get_wayland_display (lease_manager->compositor);
  MetaWaylandDrmLeaseDevice *lease_device;

  lease_device = g_rc_box_new0 (MetaWaylandDrmLeaseDevice);
  lease_device->lease_manager = lease_manager;
  lease_device->kms_device = kms_device;

  lease_device->connectors =
    g_hash_table_new_full (NULL, NULL,
                           NULL,
                           (GDestroyNotify) meta_wayland_drm_lease_connector_release);

  g_list_foreach (meta_kms_device_get_connectors (kms_device),
                  (GFunc) meta_wayland_drm_lease_device_add_connector,
                  lease_device);

  lease_device->global = wl_global_create (wayland_display,
                                           &wp_drm_lease_device_v1_interface,
                                           META_WP_DRM_LEASE_DEVICE_V1_VERSION,
                                           lease_device,
                                           lease_device_bind);

  return lease_device;
}

static void
meta_wayland_drm_lease_manager_add_device (MetaKmsDevice              *kms_device,
                                           MetaWaylandDrmLeaseManager *lease_manager)
{
  MetaWaylandDrmLeaseDevice *lease_device;

  lease_device = meta_wayland_drm_lease_device_new (lease_manager, kms_device);
  g_hash_table_insert (lease_manager->devices,
                       kms_device,
                       g_steal_pointer (&lease_device));
}

static void
meta_wayland_drm_lease_update_devices (MetaKms                    *kms,
                                       MetaWaylandDrmLeaseManager *lease_manager)
{
  GList *kms_devices;
  GList *l;
  GHashTable *new_devices;
  MetaWaylandDrmLeaseDevice *lease_device;
  GHashTableIter iter;

  new_devices = g_hash_table_new_similar (lease_manager->devices);

  kms_devices = meta_kms_get_devices (kms);
  for (l = kms_devices; l; l = l->next)
    {
      MetaKmsDevice *kms_device = l->data;

      if (!g_hash_table_steal_extended (lease_manager->devices,
                                        kms_device,
                                        NULL,
                                        (gpointer *) &lease_device))
        {
          lease_device =
            meta_wayland_drm_lease_device_new (lease_manager, kms_device);
        }

      g_hash_table_insert (new_devices, kms_device, lease_device);
    }

  g_hash_table_iter_init (&iter, lease_manager->devices);
  while (g_hash_table_iter_next (&iter, NULL, (gpointer *) &lease_device))
    {
      for (l = lease_device->lease_manager->leases; l; l = l->next)
        {
          MetaWaylandDrmLease *lease = l->data;

          if (lease->lease_device == lease_device)
            meta_wayland_drm_lease_revoke (lease);
        }

      wl_global_remove (lease_device->global);
    }

  g_hash_table_unref (lease_manager->devices);
  lease_manager->devices = new_devices;
}

static void
on_resources_changed (MetaKms                    *kms,
                      MetaKmsResourceChanges      changes,
                      MetaWaylandDrmLeaseManager *lease_manager)
{
  GHashTableIter iter;
  MetaWaylandDrmLeaseDevice *lease_device;

  if (changes != META_KMS_RESOURCE_CHANGE_FULL)
    return;

  meta_wayland_drm_lease_update_devices (kms, lease_manager);

  g_hash_table_iter_init (&iter, lease_manager->devices);
  while (g_hash_table_iter_next (&iter, NULL, (gpointer *) &lease_device))
    meta_wayland_drm_lease_device_update_connectors (lease_device);
}

MetaWaylandDrmLeaseManager *
meta_wayland_drm_lease_manager_new (MetaWaylandCompositor *compositor)
{
  MetaContext *context = meta_wayland_compositor_get_context (compositor);
  MetaBackend *backend = meta_context_get_backend (context);
  MetaBackendNative *backend_native;
  MetaKms *kms;
  MetaWaylandDrmLeaseManager *lease_manager;

  if (!META_IS_BACKEND_NATIVE (backend))
    return NULL;

  backend_native = META_BACKEND_NATIVE (backend);
  kms = meta_backend_native_get_kms (backend_native);

  lease_manager = g_new0 (MetaWaylandDrmLeaseManager, 1);
  lease_manager->compositor = compositor;
  lease_manager->kms_lease_manager = g_object_new (META_TYPE_KMS_LEASE_MANAGER,
                                                   "backend-native",
                                                   backend_native,
                                                   NULL);
  lease_manager->devices =
    g_hash_table_new_full (NULL, NULL,
                           NULL,
                           (GDestroyNotify) meta_wayland_drm_lease_device_release);

  g_list_foreach (meta_kms_get_devices (kms),
                  (GFunc) meta_wayland_drm_lease_manager_add_device,
                  lease_manager);

  g_signal_connect (kms, "resources-changed",
                    G_CALLBACK (on_resources_changed),
                    lease_manager);

  return lease_manager;
}

void
meta_wayland_drm_lease_manager_free (MetaWaylandDrmLeaseManager *lease_manager)
{
  g_clear_pointer (&lease_manager->devices, g_hash_table_unref);
  g_object_unref (lease_manager->kms_lease_manager);
  g_list_foreach (lease_manager->leases,
                  (GFunc) meta_wayland_drm_lease_release,
                  NULL);
  g_free (lease_manager);
}
